from inoon_lora_packet.packet import Packet, HexConverter


class AliveV3Packet(Packet):
    mg_per_step = {
        1: 3.91,
        2: 7.81,
        3: 15.63,
        4: 31.25,
    }

    def _field_spec(self):
        return [
            {'name': 'x',
             'bytes': 2,
             'convert': HexConverter.hex_to_int16,
             'restrict': None},

            {'name': 'y',
             'bytes': 2,
             'convert': HexConverter.hex_to_int16,
             'restrict': None},

            {'name': 'z',
             'bytes': 2,
             'convert': HexConverter.hex_to_int16,
             'restrict': None},

            {'name': 'alive_period',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'sensitivity',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'threshold',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'acc_no',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'acc_resv',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'acc_data',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'log',
             'bytes': 1,
             'bit_fields': [
                 {'name': 'axis', 'bits': 4, 'restrict': [0, 1, 2, 3]},
                 {'name': 'enable', 'bits': 4, 'restrict': [0, 1]},
             ]},

            {'name': 'log_interval',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'log_blocks',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'installed',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': [0, 1, 2]},

            {'name': 'app_fw_major',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'app_fw_minor',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'app_fw_rev',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'lora_fw_major',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'lora_fw_minor',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'lora_fw_rev',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

        ]

    @property
    def gravity_x(self):
        return self._to_gravity(self.x)

    @property
    def gravity_y(self):
        return self._to_gravity(self.y)

    @property
    def gravity_z(self):
        return self._to_gravity(self.z)

    def _to_gravity(self, raw_value):
        gravity = self.mg_per_step[self.sensitivity] * raw_value / 1000
        return round(gravity, 3)

    def __str__(self):
        log_axis_value = {
            0: 'XYZ',
            1: 'X',
            2: 'Y',
            3: 'Z',
        }

        msg = ''
        msg += 'ALIVE | '
        msg += 'Acc Conf.: {}G, {}mg | '.format(2**self.sensitivity,
                                                self.threshold)
        msg += 'Acc Intr.: {}, {} | '.format(self.acc_no, self.acc_data)
        msg += 'Acc Vals.: {}, {}, {} | '.format(self.x, self.y, self.z)
        msg += 'Log: {}({}), {}s, {}blks | '.format(
            'EN' if self.log.enable is 1 else 'DIS',
            log_axis_value[self.log.axis],
            self.log_interval, self.log_blocks)

        msg += 'Inst.: {} | '.format('Y' if self.installed is 1 else 'N')
        msg += 'Per: {}m | '.format(self.alive_period)
        msg += 'App FW: {}.{}.{} | '.format(self.app_fw_major,
                                            self.app_fw_minor,
                                            self.app_fw_rev)
        msg += 'LoRa FW: {}.{}.{}'.format(self.lora_fw_major,
                                          self.lora_fw_minor,
                                          self.lora_fw_rev)

        return msg

    @classmethod
    def encode(self, x, y, z, period, sensitivity, threshold,
               acc_no, acc_data,
               log_enabled, log_axis, log_interval, log_blocks,
               installed, app_fw_major, app_fw_minor, app_fw_rev,
               lora_fw_major, lora_fw_minor, lora_fw_rev):
        """Create alive payload for version 3 spec.
        """
        enc = ''
        enc += HexConverter.int_to_hex(x, 2, True)
        enc += HexConverter.int_to_hex(y, 2, True)
        enc += HexConverter.int_to_hex(z, 2, True)

        enc += format(period, '04x')
        enc += format(sensitivity, '02x')
        enc += format(threshold, '04x')

        enc += format(acc_no, '02x')
        enc += '00'  # Acc Resv
        enc += format(acc_data, '02x')

        enc += format(int((log_axis << 4) | log_enabled), '02x')
        enc += format(log_interval, '02x')
        enc += format(log_blocks, '02x')

        enc += format(installed, '02x')

        enc += format(app_fw_major, '02x')
        enc += format(app_fw_minor, '02x')
        enc += format(app_fw_rev, '02x')

        enc += format(lora_fw_major, '02x')
        enc += format(lora_fw_minor, '02x')
        enc += format(lora_fw_rev, '02x')

        return enc.lower()



from inoon_lora_packet.packet import Packet, HexConverter


class MRReportV3Packet(Packet):
    def _field_spec(self):
        return [
            {'name': 'mrt_protocol_version',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_battery_level',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_accumulated_runtime_duration',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_observation_time',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_reserved',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},
        ]

    def __str__(self):
        msg = ''
        msg = 'MRReport | '
        msg += 'Ver: {} | '.format(self.mrt_protocol_version)
        msg += 'BAT. : {} | '.format(self.mrt_battery_level)
        msg += 'RT. : {} | '.format(self.mrt_accumulated_runtime_duration)
        msg += 'Obsv. : {} | '.format(self.mrt_observation_time)
        msg += 'Resv. : {} | '.format(self.mrt_reserved)
        return msg

    @classmethod
    def encode(cls, version, battery, runtime_duration, observe_time, resv):
        enc = ''
        enc += format(version, '02x')
        enc += format(battery, '02x')
        enc += format(runtime_duration, '04x')
        enc += format(observe_time, '04x')
        enc += format(resv, '04x')
        return enc.lower()



from enum import IntEnum

from inoon_lora_packet.packet import Packet, HexConverter
from inoon_lora_packet.bsp.accelerometer import BMA280


class MachineState(IntEnum):
    commisioning = 0
    inactive = 1
    active = 2


class MRMeasurementV3Packet(Packet):
    def _field_spec(self):
        return [
            {'name': 'mrt_protocol_version',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'machine_state',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': [MachineState[k] for k in MachineState._member_map_]},

            {'name': 'mrt_report_period',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_report_header',
             'bytes': 1,
             'bit_fields': [
                 {'name': 'method', 'bits': 1, 'restrict': None},
                 {'name': 'overflow', 'bits': 1, 'restrict': None},
                 {'name': 'resv', 'bits': 2, 'restrict': None},
                 {'name': 'scale', 'bits': 4, 'restrict': None}
             ]},

            {'name': 'mrt_average',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_stdev',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_min',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_max',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_runtime_duration',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_operation_threshold',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_shock_threshold',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_timestamp',
             'bytes': 4,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_battery_level',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_acc_status',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'mrt_reserved',
             'bytes': 8,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

        ]

    def __str__(self):
        scale = 2 ** self.mrt_report_header.scale

        msg = 'MRMeasure | '
        msg += 'Ver: {} | '.format(self.mrt_protocol_version)
        msg += 'State: {} | '.format(
            MachineState._member_names_[self.machine_state])
        msg += 'Period: {} | '.format(self.mrt_report_period)
        msg += 'Header(Scale): {} | '.format(self.mrt_report_header.scale)
        msg += 'Avg.: {}mg | '.format(
            round(BMA280.gravity(1, self.mrt_average) * 1000 / scale, 3))
        msg += 'Std.: {}mg | '.format(
            round(BMA280.gravity(1, self.mrt_stdev) * 1000 / scale, 3))
        msg += 'Min: {}mg | '.format(
            round(BMA280.gravity(1, self.mrt_min) * 1000 / scale, 3))
        msg += 'Max: {}mg | '.format(
            round(BMA280.gravity(1, self.mrt_max) * 1000 / scale, 3))
        msg += 'RtDur.: {} | '.format(self.mrt_runtime_duration)
        msg += 'OpThr.: {} | '.format(self.mrt_operation_threshold)
        msg += 'ShkThr.: {} | '.format(self.mrt_shock_threshold)
        msg += 'Ts.: {} | '.format(self.mrt_timestamp)
        msg += 'Bat.: {} | '.format(self.mrt_battery_level)
        msg += 'Acc.: {} | '.format(self.mrt_acc_status)
        msg += 'Resv.: {}'.format(self.mrt_reserved)
        return msg

    @classmethod
    def encode(cls,
               version,
               machine_state,
               report_period,
               report_header,
               min_val,
               max_val,
               avg,
               std,
               runtime_duration,
               operation_threshold,
               shock_threshold,
               timestamp,
               battery,
               acc_status,
               resv):
        enc = ''
        enc += format(version, '02x')
        enc += format(machine_state, '02x')
        enc += format(report_period, '02x')
        enc += format(report_header, '02x')
        enc += format(avg, '04x')
        enc += format(std, '04x')
        enc += format(min_val, '04x')
        enc += format(max_val, '04x')
        enc += format(runtime_duration, '04x')
        enc += format(operation_threshold, '04x')
        enc += format(shock_threshold, '04x')
        enc += format(timestamp, '08x')
        enc += format(battery, '02x')
        enc += format(acc_status, '02x')
        enc += format(resv, '016x')

        return enc.lower()



import abc
import json
from enum import IntEnum


class InvalidPacketError(Exception):
    def __init__(self):
        super(self.__class__, self).__init__('Invalid RAW Packet.')


class NotSupportedPacketError(Exception):
    def __init__(self):
        super(self.__class__, self).__init__('Not Supported packet')


class RestrictedValueError(Exception):
    def __init__(self):
        super(self.__class__, self).__init__('Restricted Value Occured')


class HexConverter():
    # TODO: Convert bugs..
    @classmethod
    def _hex_to_sint(cls, hexstr, size):
        value = int(hexstr[-2*size:], 16)
        if value & (0x1 << (size * 8 - 1)):
            bit_mask = 0
            for _ in range(0, size):
                bit_mask = (bit_mask << 8) | 0xFF

            absolute = (value ^ bit_mask) + 1
            convert_value = -absolute
        else:
            convert_value = value
        return convert_value

    @classmethod
    def hex_to_int8(cls, hexstr):
        return HexConverter._hex_to_sint(hexstr, 1)

    @classmethod
    def hex_to_int16(cls, hexstr):
        return HexConverter._hex_to_sint(hexstr, 2)

    @classmethod
    def hex_to_uint(cls, hexstr):
        return int(hexstr, 16)

    @classmethod
    def int_to_hex(cls, value, byte_len, is_signed):
        mask = 0
        for _ in range(0, byte_len):
            mask = (mask << 8) | 0xFF

        if is_signed is True and value < 0:
            hex_val = ((abs(value) ^ mask) & mask) + 1
        else:
            hex_val = value

        result_len = byte_len * 2
        hex_str = '{:X}'.format(hex_val)
        remain_len = result_len - len(hex_str)

        padding = ''
        for _ in range(0, remain_len):
            padding += '0'

        return padding + hex_str


class BitField():
    pass


class DeviceType(IntEnum):
    mgi = 2
    mgi_100n = 3


class PacketType(IntEnum):
    alive = 1
    event = 2
    error = 3
    ack = 4
    notice = 5
    data_log = 6
    report = 7
    acc_wave = 8
    inclination = 9
    mr_measurement = 10
    mr_report = 11


class RequestType(IntEnum):
    none = 0
    sync = 1
    config = 2


class EventType(IntEnum):
    high_g = 1
    collapse = 8


class ErrorType(IntEnum):
    lora_over_re_tx = 1
    lora_join_fail = 2
    lora_rx = 3
    lora_nok = 4
    lora_tx_inf = 5
    lora_skip_high_g = 6
    lora_flash_store = 7
    lora_flash_load = 8
    lora_flash_clear = 9
    lora_flash_update = 10


class AckType(IntEnum):
    ext_mgmt = 0x00,
    dev_reset = 0x80,
    per_change = 0x81,
    immediate = 0x82


class NoticeType(IntEnum):
    power_up = 1,
    power_off = 2,
    reset = 3,
    setup = 4,
    test_result = 5


class LogPosition(IntEnum):
    start = 0,
    intermid = 1,
    finish = 2


class Packet(abc.ABC):
    def __init__(self, raw_packet):
        self.raw_packet = raw_packet
        self._parse(raw_packet)

    @abc.abstractmethod
    def _field_spec(self):
        pass

    def _parse(self, raw_packet):
        last_idx = 0
        for i, spec in enumerate(self._field_spec()):
            start = last_idx * 2

            length = int(spec['bytes'])
            end = start + 2*length

            raw_value = raw_packet[start:end]

            # TODO: Simplify code.
            if 'bit_fields' in spec:
                setattr(self, spec['name'], BitField())
                new_attr = getattr(self, spec['name'])

                bits = 8
                for i, bit_spec in enumerate(spec['bit_fields']):
                    bit_mask = 0
                    for i in range(0, bit_spec['bits']):
                        bit_mask = bit_mask << 1
                        bit_mask |= 0x1

                    bit_mask = bit_mask << (bits - bit_spec['bits'])
                    bit_value = (int(raw_value, 16) & bit_mask)
                    bit_value = bit_value >> (bits - bit_spec['bits'])

                    self._valid_restrict(bit_spec['restrict'], bit_value)

                    setattr(new_attr, bit_spec['name'], bit_value)

                    bits -= bit_spec['bits']
            else:
                try:
                    convert_val = spec['convert'](raw_value)
                except ValueError:
                    raise InvalidPacketError

                self._valid_restrict(spec['restrict'], convert_val)
                setattr(self, spec['name'], convert_val)

            last_idx += length

    def _valid_restrict(self, restrict, value):
        if restrict is not None and value not in restrict:
            raise RestrictedValueError

    def json(self):
        return json.dumps(self._dump(self.__dict__))

    def _dump(self, dict_info):
        result = {}

        for key, value in dict_info.items():
            if type(value).__name__ in __builtins__:
                result[key] = value
            else:
                result[key] = self._dump(value.__dict__)

        return result


class PacketHeaderV2(Packet):
    def _field_spec(self):
        return [
            {'name': 'ver',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': [2]},

            {'name': 'resv',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'id',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'dev',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': [1, 2]},

            {'name': 'packet',
             'bytes': 1,
             'bit_fields': [
                 {'name': 'type', 'bits': 4, 'restrict': [1, 2, 3, 4, 5, 6]},
                 {'name': 'req', 'bits': 4, 'restrict': [0, 1, 2]},
             ]},

            {'name': 'battery',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': range(0, 101)},

            {'name': 'temperature',
             'bytes': 1,
             'convert': HexConverter.hex_to_int8,
             'restrict': range(-30, 80)},
        ]

    def __init__(self, raw_packet):
        super(self.__class__, self).__init__(raw_packet)

    def __str__(self):
        msg = ''
        msg += 'V{} | '.format(self.ver)
        msg += 'ID: {} | '.format(self.id)
        msg += 'BAT.: {} | '.format(self.battery)
        msg += 'TEMP.: {}'.format(self.temperature)
        return msg

    @classmethod
    def encode(cls, id, dev_type, packet_type, req_type, bat, temp):
        enc = ''
        enc += format(2, '02x')
        enc += '00'
        enc += format(id, '02x')
        enc += format(dev_type, '04x')
        enc += format(packet_type, 'x')
        enc += format(req_type, 'x')
        enc += format(bat, '02x')
        enc += HexConverter.int_to_hex(temp, 1, True)
        return enc.lower()


class PacketHeaderV3(Packet):
    packet_type_enums = [PacketType[k] for k in PacketType._member_map_]

    def _field_spec(self):
        return [
            {'name': 'ver',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': [3]},

            {'name': 'dev_type',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': [DeviceType[k] for k in DeviceType._member_map_]},

            {'name': 'pckt_seq',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'bat',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': range(0, 101)},

            {'name': 'temp',
             'bytes': 1,
             'convert': HexConverter.hex_to_int8,
             'restrict': range(-30, 128)},

            {'name': 'lora_err',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

            {'name': 'rssi',
             'bytes': 1,
             'convert': HexConverter.hex_to_int8,
             'restrict': None},

            {'name': 'packet',
             'bytes': 1,
             'bit_fields': [
                 {'name': 'type',
                  'bits': 4,
                  'restrict': PacketHeaderV3.packet_type_enums},
                 {'name': 'req', 'bits': 4, 'restrict': [0, 1, 2]},
             ]},

            {'name': 'resv',
             'bytes': 2,
             'convert': HexConverter.hex_to_uint,
             'restrict': None},

        ]

    def __init__(self, raw_packet):
        super(self.__class__, self).__init__(raw_packet)

    @property
    def battery(self):
        """For V2 compatibility"""
        return self.bat

    @property
    def temperature(self):
        """For V2 compatibility"""
        return self.temp

    def __str__(self):
        msg = ''
        msg += 'V{} | '.format(self.ver)
        msg += 'Dev.: {} | '.format(DeviceType(self.dev_type)._name_)
        msg += 'Seq.: {} | '.format(self.pckt_seq)
        msg += 'Bat.: {} | '.format(self.bat)

        if self.temp == 0x7F:
            msg += 'BMA Err'
        else:
            msg += 'Temp.: {} | '.format(self.temp)

        msg += 'Err.: {} | '.format(self.lora_err)
        msg += 'Type: {}, {} | '.format(PacketType(self.packet.type)._name_,
                                        RequestType(self.packet.req)._name_)
        msg += 'RSSI.: {} | '.format(self.rssi)
        msg += 'Resv.: {}'.format(self.resv)
        return msg

    @classmethod
    def encode(cls, dev_type, seq, bat, temp, lora_err, rssi,
               payload_type, payload_req, resv):
        enc = ''
        enc += '03'
        enc += format(dev_type, '02x')
        enc += format(seq, '02x')
        enc += format(bat, '02x')
        enc += HexConverter.int_to_hex(temp, 1, True)
        enc += format(lora_err, '02x')
        enc += HexConverter.int_to_hex(rssi, 1, True)

        enc += format(payload_type, 'x')
        enc += format(payload_req, 'x')
        enc += format(resv, '04x')
        return enc.lower()



from inoon_lora_packet.packet import (Packet, InvalidPacketError, HexConverter)


class AccWaveV3Packet(Packet):
    axis_ctrl = {
        0: ['x', 'y', 'z'],
        1: ['x'],
        2: ['y'],
        3: ['z']
    }

    def _field_spec(self):
        return [
            {'name': 'ctrl',
             'bytes': 1,
             'bit_fields': [
                 {'name': 'resv', 'bits': 2, 'restrict': None},
                 {'name': 'axis', 'bits': 2, 'restrict': [0, 1, 2, 3]},
                 {'name': 'id', 'bits': 4, 'restrict': None}
             ]},
            {'name': 'seq',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': range(0, 256)}
        ]

    def __init__(self, raw_packet):
        super(self.__class__, self).__init__(raw_packet)

        log_packet = raw_packet[4:]
        axises = self.axis_ctrl[self.ctrl.axis]

        if len(log_packet) % (len(axises) * 4) != 0:
            raise InvalidPacketError

        for axis in axises:
            setattr(self, axis, [])

        frame_len = len(axises) * 4
        for i in range(0, len(log_packet), frame_len):
            frame = log_packet[i:i+frame_len]
            for i, axis in enumerate(axises):
                value = HexConverter.hex_to_int16(frame[i*4:(i+1)*4])
                getattr(self, axis).append(value)

    def __str__(self):
        msg = ''
        msg += 'WAVE | '
        msg += 'Axis: {} | '.format(' '.join(self.axis_ctrl[self.ctrl.axis]))
        msg += 'ID: {} | '.format(self.ctrl.id)
        msg += 'Seq: {}'.format(self.seq)
        return msg

    @classmethod
    def encode(cls, pckt_id, seq, axis, values):
        """values parameter is dict type.
        """
        enc = ''

        ctrl = (axis << 4) | pckt_id
        enc += '{:02X}'.format(ctrl)
        enc += '{:02X}'.format(seq)

        if axis == 0:
            acc_values = list(zip(values['x'], values['y'], values['z']))

            for value in acc_values:
                enc += HexConverter.int_to_hex(value[0], 2, True)
                enc += HexConverter.int_to_hex(value[1], 2, True)
                enc += HexConverter.int_to_hex(value[2], 2, True)
        else:
            axis_name = ''
            if axis == 1:
                axis_name = 'x'
            elif axis == 2:
                axis_name = 'y'
            else:
                axis_name = 'z'

            for value in values[axis_name]:
                enc += HexConverter.int_to_hex(value, 2, True)

        return enc.lower()



from inoon_lora_packet.packet import Packet, HexConverter, AckType


class AckV3Packet(Packet):
    def _field_spec(self):
        return [
            {'name': 'ack_type',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': [AckType[k] for k in AckType._member_map_]
             },
        ]

    def __str__(self):
        code_desc = {
            AckType.ext_mgmt: 'ExtDevMgmt',
            AckType.dev_reset: 'DevReset',
            AckType.per_change: 'RepPerChange',
            AckType.immediate: 'RepImmediate'
        }

        msg = 'ACK | '
        msg += 'TYPE: {}({}) | '.format(code_desc[self.ack_type],
                                        self.ack_type)
        return msg

    @classmethod
    def encode(cls, ack_type):
        return format(ack_type, '02x').lower()



from .packet import InvalidPacketError, NotSupportedPacketError
from .v2.parser import Ver2Parser
from .v3.parser import Ver3Parser


class ParserFactory():
    @classmethod
    def get_parser(cls, version):

        if version is 1:
            return Ver2Parser()
        elif version is 2:
            return Ver2Parser()
        elif version is 3:
            return Ver3Parser()
        else:
            raise NotSupportedPacketError


class PacketParser():
    @classmethod
    def parse(cls, raw_packet):
        try:
            version = int(raw_packet[:2], 16)
            parser = ParserFactory.get_parser(version=version)
            return parser.parse(raw_packet=raw_packet)
        except ValueError:
            raise InvalidPacketError



from inoon_lora_packet.packet import Packet, HexConverter


class InclinationV3Packet(Packet):
    def _field_spec(self):
        return [
            {'name': 'x_degree',
             'bytes': 2,
             'convert': HexConverter.hex_to_int16,
             'restrict': None},
            {'name': 'y_degree',
             'bytes': 2,
             'convert': HexConverter.hex_to_int16,
             'restrict': None},
            {'name': 'z_degree',
             'bytes': 2,
             'convert': HexConverter.hex_to_int16,
             'restrict': None},
        ]

    def __init__(self, raw_packet):
        super(self.__class__, self).__init__(raw_packet)

    def __str__(self):
        msg = 'INCLINATION | '
        msg += 'Degree Diff.: {}, {}, {}'.format(self.x_degree / 100.0,
                                                 self.y_degree / 100.0,
                                                 self.z_degree / 100.0)

        return msg

    @classmethod
    def encode(cls, x, y, z):
        enc = ''
        enc += '{}'.format(HexConverter.int_to_hex(x, 2, True))
        enc += '{}'.format(HexConverter.int_to_hex(y, 2, True))
        enc += '{}'.format(HexConverter.int_to_hex(z, 2, True))
        return enc.lower()



from inoon_lora_packet.packet import (Packet, NoticeType,
                                      HexConverter, InvalidPacketError)


# TODO: Refactoring this...

class NoticeV3Packet(Packet):
    err_desc = {
        0x00: 'No reason.',
        0x01: 'Exceed Alive period.',
        0x02: 'LoRa TX fail.',
        0x04: 'LoRa join fail.',
        0x05: 'Watchdog,',
        0x06: 'No Ack for PowerOff notice.',
        0x07: 'LoRa PowerOff Command.',
        0x08: 'BLE PowerOff.',
        0x09: 'Low Battery.',
        0x0A: 'Exceed Install request.',
        0x0B: 'RESET_CONFIG.',
        0x0C: 'PowefOffUninstallCommand',
        0x0D: 'PowerOffUninstallByResv.',
        0x0E: 'PowerOffDeviceUpsideDown',
        0x0F: 'PowerOffUninstallTimeout',
        0x81: 'Factory reset.',
        0x82: 'Upside down.',
        0x83: 'Watchdod changed.',
        0x84: 'LoRa reset.',
        0x85: 'BLE reset',
        0x86: 'Factory reset',
        0x87: 'LoRa SKT reset',
        0x88: 'LoRa FW update reset.',
        0x89: 'Init config reset',
        0x8a: 'Power up notice fail reset',
        0x8b: 'Report params change reset',
        0x8c: 'Inclination params change reset',
        0x8d: 'Runtime config change reset',
        0x8e: 'ReTx filed reset',
        0x8f: 'Infinite reset test',
        0x90: 'Wave timeout reset',
        0x91: 'LoRa join fail reset',
        0x92: 'NOK reset',
        0x93: 'Abnormal reset',
        0x94: 'Alive Failed reset',
    }

    setup_desc = {
        0: 'Uninstall',
        1: 'Install',
        2: 'ReqInstall'
    }

    def _field_spec(self):
        try:
            notice_type = int(self.raw_packet[:2], 16)

            if notice_type == NoticeType.power_up:
                return PowerUpV3Packet.spec
            elif notice_type == NoticeType.power_off:
                return PowerOffV3Packet.spec
            elif notice_type == NoticeType.setup:
                return SetupV3Packet.spec
            elif notice_type == NoticeType.test_result:
                return TestResultV3Packet.spec

            raise InvalidPacketError
        except Exception:
            raise InvalidPacketError

    def __str__(self):
        msg = None
        if self.type == NoticeType.power_up:
            msg = self.__power_up_log_msg()
        elif self.type == NoticeType.power_off:
            msg = self.__power_off_log_msg()
        elif self.type == NoticeType.setup:
            msg = self.__setup_msg()
        elif self.type == NoticeType.test_result:
            msg = self.__test_result_log_msg()

        return msg

    def __power_up_log_msg(self):
        msg = 'ON | '
        msg += 'Reset({}): '.format(self.reason)
        if self.reason & 0x08 == 0x08:
            msg += 'CPU '
        if self.reason & 0x04 == 0x04:
            msg += 'SW '
        if self.reason & 0x02 == 0x02:
            msg += 'WD '
        if self.reason & 0x01 == 0x01:
            msg += 'nRST'
        if self.reason is 0:
            msg += '--'
        msg += ' | '

        msg += 'Err({}): {}'.format(self.error, self.err_desc[self.error])

        return msg

    def __power_off_log_msg(self):
        msg = 'OFF | '
        msg += 'Reason({}): {}'.format(self.reason, self.err_desc[self.reason])

        return msg

    def __test_result_log_msg(self):
        msg = 'TEST_RESULT | '
        msg += 'Enabled' if self.enable == 1 else 'Disabled'
        msg += 'Command {} |'.format(self.command)
        msg += 'CountLimit {} |'.format(self.count_limit)
        msg += 'MoveToNextFlag {} |'.format(self.move_to_next_flag)
        msg += 'CurrentCount {} |'.format(self.current_count)
        msg += 'running_flag {} |'.format(self.running_flag)
        msg += 'resv {} |'.format(self.resv)

    def __setup_msg(self):
        return 'SETUP | {} -> {}'.format(self.previous, self.current)


class PowerUpV3Packet():
    spec = [
        {'name': 'type',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [NoticeType.power_up, NoticeType.power_off]},

        {'name': 'len',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [4]},

        {'name': 'reason',
         'bytes': '1',
         'convert': HexConverter.hex_to_uint,
         'restrict': None},

        {'name': 'resv',
         'bytes': '2',
         'convert': HexConverter.hex_to_uint,
         'restrict': None},

        {'name': 'error',
         'bytes': '1',
         'convert': HexConverter.hex_to_uint,
         'restrict': [key for key in NoticeV3Packet.err_desc.keys()]},

    ]

    @classmethod
    def encode(cls, reason, error):
        enc_val = ''
        enc_val += format(1, '02x')
        enc_val += format(4, '02x')
        enc_val += format(reason, '02x')
        enc_val += '0000'
        enc_val += format(error, '02x')
        return enc_val.lower()


class PowerOffV3Packet():
    spec = [
        {'name': 'type',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [NoticeType.power_up, NoticeType.power_off]},

        {'name': 'len',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [1]},

        {'name': 'reason',
         'bytes': '1',
         'convert': HexConverter.hex_to_uint,
         'restrict': [key for key in NoticeV3Packet.err_desc.keys()]},
    ]

    @classmethod
    def encode(cls, reason):
        enc_val = ''
        enc_val += format(2, '02x')
        enc_val += format(1, '02x')
        enc_val += format(reason, '02x')
        return enc_val.lower()


class SetupV3Packet(Packet):
    spec = [
        {'name': 'type',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': None},

        {'name': 'len',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [2]},

        {'name': 'current',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [0, 1, 2]},

        {'name': 'previous',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [0, 1, 2]},
    ]

    @classmethod
    def encode(cls, prev_install, current_install):
        enc_val = ''
        enc_val += format(NoticeType.setup, '02x')
        enc_val += format(2, '02x')
        enc_val += format(current_install, '02x')
        enc_val += format(prev_install, '02x')
        return enc_val.lower()


class TestResultV3Packet(Packet):
    spec = [
        {'name': 'type',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': None},

        {'name': 'len',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [8]},

        {'name': 'enable',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [0, 1]},

        {'name': 'command',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': range(1, 7)},

        {'name': 'count_limit',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': range(0, 256)},

        {'name': 'move_to_next_flag',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [0, 1]},

        {'name': 'current_count',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': range(0, 256)},

        {'name': 'running_flag',
         'bytes': 1,
         'convert': HexConverter.hex_to_uint,
         'restrict': [0, 1]},

        {'name': 'resv',
         'bytes': 2,
         'convert': HexConverter.hex_to_uint,
         'restrict': None},
    ]

    @classmethod
    def encode(cls, enable, command, count_limit, move_to_next_flag,
               current_count, running_flag, resv):
        enc_val = ''
        enc_val += format(NoticeType.test_result, '02x')
        enc_val += format(8, '02x')
        enc_val += format(enable, '02x')
        enc_val += format(command, '02x')
        enc_val += format(count_limit, '02x')
        enc_val += format(move_to_next_flag, '02x')
        enc_val += format(current_count, '02x')
        enc_val += format(running_flag, '02x')
        enc_val += format(resv, '04x')
        return enc_val.lower()



from inoon_lora_packet.packet import Packet, HexConverter, AckType


class AckV2Packet(Packet):
    def _field_spec(self):
        return [
            {'name': 'ack_type',
             'bytes': 1,
             'convert': HexConverter.hex_to_uint,
             'restrict': [AckType[k] for k in AckType._member_map_]},

            {'name': 'rssi',
             'bytes': 1,
             'convert': HexConverter.hex_to_int8,
             'restrict': None},
        ]

    def __str__(self):
        msg = 'ACK | '
        msg += 'TYPE: {} | '.format(self.ack_type)
        msg += 'RSSI: {}'.format(self.rssi)
        return msg

    @classmethod
    def encode(cls, ack_type, rssi):
        enc = ''
        enc += format(ack_type, '02x')
        enc += HexConverter.int_to_hex(rssi, 1, True)
        return enc



from inoon_lora_packet.packet import (PacketType, PacketHeaderV2,
                                      InvalidPacketError,
                                      NotSupportedPacketError)
from inoon_lora_packet.base import base_parser
from .alive import AliveV2Packet
from .notice import NoticeV2Packet
from .data_log import DataLogV2Packet
from .event import EventV2Packet
from .ack import AckV2Packet
from .error import ErrorV2Packet


class Ver2Parser(base_parser.Parser):
    @property
    def header_length(self):
        return 16

    def parse(self, raw_packet):
        if len(raw_packet) < self.header_length:
            raise InvalidPacketError

        header = PacketHeaderV2(raw_packet)
        raw_payload = raw_packet[self.header_length:]
        packet = None

        if header.packet.type == PacketType.alive:
            packet = AliveV2Packet(raw_payload)
        elif header.packet.type == PacketType.notice:
            packet = NoticeV2Packet(raw_payload)
        elif header.packet.type == PacketType.data_log:
            packet = DataLogV2Packet(raw_payload)
        elif header.packet.type == PacketType.event:
            packet = EventV2Packet(raw_payload)
        elif header.packet.type == PacketType.ack:
            packet = AckV2Packet(raw_payload)
        elif header.packet.type == PacketType.error:
            packet = ErrorV2Packet(raw_payload)
        else:
            raise NotSupportedPacketError

        return header, packet
